@@ -9,7 +9,7 @@ import sqlite3
 import tempfile
 from contextlib import contextmanager
 from inspect import isgeneratorfunction
-from itertools import chain, groupby
+from itertools import chain, groupby, product
 from pathlib import Path, PurePath
 
 
@@ -69,6 +69,14 @@ class _InstructionContainer:
         })
         return self
 
+    # takes a pre-listed group (each group is an iterator itself)
+    def _map(self, fn):
+        self._insts.append({
+            'cmd': '_map',
+            'fn': fn
+        })
+        return self
+
     def by(self, *cols):
         self._insts.append({
             'cmd': 'by',
@@ -84,6 +92,23 @@ class _InstructionContainer:
         })
         return self
 
+    # takes a pre-listed group (each group is an iterator itself)
+    def _merge(self, fn, other):
+        self._insts.append({
+            'cmd': 'merge',
+            'fn': fn,
+            'other': other
+        })
+        return self
+
+    def _merge_with_columns(self, fn, other):
+        self._insts.append({
+            'cmd': '_merge_with_columns',
+            'fn': fn,
+            'other': other
+        })
+        return self
+
     def concat(self, other):
         self._insts.append({
             'cmd': 'concat',
@@ -97,12 +122,7 @@ class _InstructionContainer:
                 if pred(r):
                     yield r
             return gen
-
-        self._insts.append({
-            'cmd': 'map',
-            'fn': predicate2gen(pred)
-        })
-        return self
+        return self.map(predicate2gen(pred))
 
     def update(self, **kwargs):
         def updatefn(r):
@@ -110,72 +130,89 @@ class _InstructionContainer:
                 r.update({k: v(r)})
             yield r
 
-        self._insts.append({
-            'cmd': 'map',
-            'fn': updatefn
-        })
-        return self
+        return self.map(updatefn)
+
+    def fold(self, **kwargs):
+        def foldfn(rs):
+            r0 = rs[0]
+            for k, v in kwargs.items():
+                r0.update({k: v(rs)})
+            yield r0
+
+        return self.map(foldfn)
 
     def select(self, *cols):
         def selectfn(r):
             return {col: r[col] for col in cols}
 
-        self._insts.append({
-            'cmd': 'map',
-            'fn': selectfn
-        })
-        return self
+        return self.map(selectfn)
 
-    def join(self, cols, other):
-        def innerjoin_fn(rs1, rs2):
+    def deselect(self, *cols):
+        cols = set(cols)
+
+        def deselectfn(r):
+            return {k: v for k, v in r.items() if k not in cols}
+
+        return self.map(deselectfn)
+
+    def inner_join(self, other):
+        def innerjoin_fn(rs1, rs2, left, center, right):
             if rs1 != [] and rs2 != []:
-                for r1 in rs1:
-                    for r2 in rs2:
-                        r1.update(r2)
-                        yield {col: r1.get(col, None) for col in cols}
+                for r1, r2 in product(rs1, rs2):
+                    r1.update(r2)
+                    yield r1
+        return self._merge_with_columns(innerjoin_fn, other)
 
-        self._insts.append({
-            'cmd': 'merge',
-            'fn': innerjoin_fn,
-            'other': other
-        })
-        return self
-
-    def left_join(self, cols, other):
-        def leftjoin_fn(rs1, rs2):
+    def left_join(self, other):
+        def leftjoin_fn(rs1, rs2, left, center, right):
             if rs1 != [] and rs2 != []:
-                for r1 in rs1:
-                    for r2 in rs2:
-                        r1.update(r2)
-                        yield {col: r1.get(col, None) for col in cols}
+                for r1, r2 in product(rs1, rs2):
+                    r1.update(r2)
+                    yield r1
+
             elif rs1 != [] and rs2 == []:
                 for r1 in rs1:
-                    yield {col: r1.get(col, None) for col in cols}
+                    for col in right:
+                        r1[col] = None
+                    yield r1
 
-        self._insts.append({
-            'cmd': 'merge',
-            'fn': leftjoin_fn,
-            'other': other
-        })
-        return self
+        return self._merge_with_columns(leftjoin_fn, other)
 
-    def right_join(self, cols, other):
-        def rightjoin_fn(rs1, rs2):
+    def right_join(self, other):
+        def rightjoin_fn(rs1, rs2, left, center, right):
             if rs1 != [] and rs2 != []:
-                for r2 in rs2:
-                    for r1 in rs1:
-                        r2.update(r1)
-                        yield {col: r2.get(col, None) for col in cols}
+                for r1, r2 in product(rs1, rs2):
+                    r2.update(r1)
+                    yield r2
+
             elif rs1 == [] and rs2 != []:
                 for r2 in rs2:
-                    yield {col: r2.get(col, None) for col in cols}
+                    for col in left:
+                        r2[col] = None
+                    yield r2
 
-        self._insts.append({
-            'cmd': 'merge',
-            'fn': rightjoin_fn,
-            'other': other
-        })
-        return self
+        return self._merge_with_columns(rightjoin_fn, other)
+
+    def full_join(self, other):
+        def fulljoin_fn(rs1, rs2, left, center, right):
+            if rs1 != [] and rs2 != []:
+                for r1, r2 in product(rs1, rs2):
+                    r1.update(r2)
+                    yield r1
+
+            elif rs1 != [] and rs2 == []:
+                for r1 in rs1:
+                    for col in right:
+                        r1[col] = None
+                    yield r1
+
+            elif rs1 == [] and rs2 != []:
+                for r2 in rs2:
+                    d = {col: None for col in left}
+                    d.update(r2)
+                    yield d
+
+        return self._merge_with_columns(fulljoin_fn, other)
 
     def iter(self):
         insts = self._insts
@@ -221,7 +258,13 @@ class _InstructionContainer:
                         for _, rs in pgen():
                             yield from fn(list(rs))
 
-                    return gen2 if 'by' in pinst['cmd'] else gen1
+                    def _gen2():
+                        for _, rs in pgen():
+                            yield from fn(rs)
+
+                    if not 'by' in pinst['cmd']:
+                        return gen1
+                    return _gen2 if inst['cmd'] == "_map" else gen2
 
                 genfns.append(buildgen(pinst, inst))
 
@@ -255,8 +298,44 @@ class _InstructionContainer:
                     pgen = genfns[-1]
 
                     def gen():
+
+                        seq1 = _itered_group_to_list(pgen())
+                        seq2 = _itered_group_to_list(_inst2iter(inst['other']))
+                        yield from _step(fn, seq1, seq2)
+
+                    def _gen():
                         yield from _step(fn, pgen(), _inst2iter(inst['other']))
 
+                    return _gen if inst['cmd'] == '_merge' else gen
+
+                genfns.append(buildgen(inst))
+
+            elif inst['cmd'] == '_merge_with_columns':
+                if not 'by' in pinst['cmd']:
+                    raise ValueError("Must be grouped by columns before merge")
+
+                def buildgen(inst):
+                    fn = _fn2gen(inst['fn'])
+                    pgen = genfns[-1]
+
+                    (_, rs1), seq1 = _spyg(pgen())
+                    (_, rs2), seq2 = _spyg(_inst2iter(inst['other']))
+
+                    cols1 = list(next(rs1))
+                    cols2 = list(next(rs2))
+
+                    left = [c for c in cols1 if c not in cols2]
+                    # common columns
+                    center = [c for c in cols1 if c in cols2]
+                    right = [c for c in cols2 if c not in cols1]
+
+                    def _trans(fn):
+                        def innerfn(rs1, rs2):
+                            yield from fn(rs1, rs2, left, center, right)
+                        return innerfn
+
+                    def gen():
+                        yield from _step(_trans(fn), seq1, seq2)
                     return gen
 
                 genfns.append(buildgen(inst))
@@ -386,30 +465,45 @@ def _step(fn, krs1, krs2):
         k2, rs2 = next(krs2)
         while True:
             if k1 == k2:
-                yield from fn(list(rs1), list(rs2))
+                yield from fn(rs1, rs2)
                 k1 = k2 = empty
                 k1, rs1 = next(krs1)
                 k2, rs2 = next(krs2)
             elif k1 < k2:
-                yield from fn(list(rs1), [])
+                yield from fn(rs1, [])
                 k1 = empty
                 k1, rs1 = next(krs1)
             else:
-                yield from fn([], list(rs2))
+                yield from fn([], rs2)
                 k2 = empty
                 k2, rs2 = next(krs2)
     except StopIteration:
         # unconsumed
         if k1 is not empty:
-            yield from fn(list(rs1), [])
+            yield from fn(rs1, [])
         if k2 is not empty:
-            yield from fn([], list(rs2))
+            yield from fn([], rs2)
 
         for _, rs1 in krs1:
-            yield from fn(list(rs1), [])
+            yield from fn(rs1, [])
         for _, rs2 in krs2:
-            yield from fn([], list(rs2))
+            yield from fn([], rs2)
 
 
 def _inst2iter(obj):
     return obj.iter() if isinstance(obj, _InstructionContainer) else iter(obj)
+
+
+# itertools.groupby group sequence where each group is an iterator itself,
+# turn it into a list
+def _itered_group_to_list(seq):
+    for k, v in seq:
+        yield k, list(v)
+
+
+# spying a grouped seq
+def _spyg(iterable):
+    it = iter(iterable)
+    k, rs = next(it)
+    rs1 = list(rs)
+    return (k, iter(rs1)), chain([(k, iter(rs1))], it)
