@@ -232,22 +232,50 @@ class TestDerivs(unittest.TestCase):
         r = next(conn['t1_1'].select('col1', 'col3').iter())
         self.assertEqual(len(r), 2)
 
+        # test deselect
+        r = next(conn['t1_1'].deselect('col1', 'col3').iter())
+        self.assertEqual(len(r), 1)
+
     def test_join(self):
         conn = Conn(self.dbfile)
         conn['t1'] = table1
         conn['t2'] = table2
 
-        conn['t1_1'] = conn['t1'].by('col1').join(['col1', 'col2', 'col3'],
-                                                  conn['t2'].by('col1'))
+        conn['t1_1'] = conn['t1'].by('col1').inner_join(conn['t2'].by('col1'))
 
         self.assertEqual(len(conn['t1_1'].list()), 3)
 
-        conn['t1_2'] = conn['t1'].by('col1').left_join(['col1', 'col2', 'col3'],
-                                                       conn['t2'].by('col1'))
+        conn['t1_2'] = conn['t1'].by('col1').left_join(conn['t2'].by('col1'))
 
         self.assertEqual(len(conn['t1_2'].list()), 8)
 
-        conn['t1_3'] = conn['t1'].by('col1').right_join(['col1', 'col2', 'col3'],
-                                                        conn['t2'].by('col1'))
+        conn['t1_3'] = conn['t1'].by('col1').right_join(conn['t2'].by('col1'))
 
         self.assertEqual(len(conn['t1_3'].list()), 5)
+
+        conn['t1_4'] = conn['t1'].by('col1').full_join(conn['t2'].by('col1'))
+
+        self.assertEqual(conn['t1_4'].list(), [
+            {'col1': 'a', 'col2': 4, 'col3': None},
+            {'col1': 'a', 'col2': 5, 'col3': None},
+            {'col1': 'b', 'col2': 1, 'col3': 1},
+            {'col1': 'c', 'col2': 3, 'col3': None},
+            {'col1': 'c', 'col2': 4, 'col3': None},
+            {'col1': 'c', 'col2': 7, 'col3': None},
+            {'col1': 'd', 'col2': 2, 'col3': 3},
+            {'col1': 'd', 'col2': 2, 'col3': 2},
+            {'col1': 'e', 'col2': None, 'col3': 3},
+            {'col1': 'e', 'col2': None, 'col3': 4}]
+        )
+
+    def test_fold(self):
+        conn = Conn(self.dbfile)
+        conn['t1'] = table1
+
+        conn['t1_1'] = conn['t1'].by('col1').fold(
+            tot=lambda rs: sum(r['col2'] for r in rs))
+        self.assertEqual(conn['t1_1'].list(),
+                         [{'col1': 'a', 'col2': 4, 'tot': 9},
+                          {'col1': 'b', 'col2': 1, 'tot': 1},
+                          {'col1': 'c', 'col2': 3, 'tot': 14},
+                          {'col1': 'd', 'col2': 2, 'tot': 2}])
